import toml
import numpy as np
from scipy.spatial.transform import Rotation
import logging as log
import pandas as pd

from .transform import Transform


class TransformSet:
    def __init__(self, cfg_file: str) -> None:

        self.frame_data = toml.load(cfg_file)

        # Save names of transforms
        self.frame_names = []
        for frame_name in self.frame_data.keys():
            self.frame_names.append(frame_name)

        if 'base' not in self.frame_names:
            log.error(f"TransformSet - No frame is marked as base frame. Please mark one of the frames as 'base'.")
            return

        # Convert dictionary parameters to a list of transformations
        self.transformations = []
        valid_rotation_types = ['euler', 'quaternion', 'rotvec', 'matrix', 'rodrigues']
        for frame_name in self.frame_data.keys():
            new_transf = Transform(name=frame_name, orig='base', dest=frame_name)
            new_transf.translation = self.frame_data[frame_name]['translation']

            degree_opt = 'degree' in self.frame_data[frame_name]['orientation_units']
            orientation_type = self.frame_data[frame_name]['orientation_type']
            orientation_value = self.frame_data[frame_name]['orientation']

            if orientation_type not in valid_rotation_types:
                log.error(f"TransformSet - Invalid rotation type: {orientation_type}. Rotation type must be: {valid_rotation_types}")
                continue
            elif orientation_type == 'euler':
                new_transf.rotation = Rotation.from_euler('xyz', orientation_value, degrees=degree_opt)
            elif orientation_type == 'quaternion':
                new_transf.rotation = Rotation.from_quat(orientation_value)
            elif orientation_type == 'rotvec':
                new_transf.rotation = Rotation.from_rotvec(orientation_value, degrees=degree_opt)
            elif orientation_type == 'matrix':
                new_transf.rotation = Rotation.from_matrix(orientation_value)
            elif orientation_type == 'rodrigues':
                new_transf.rotation = Rotation.from_mrp(orientation_value)

            self.transformations.append(new_transf)


    def change_frame(self, input, from_frame: str, to_frame: str):
        # Create compound transformation
        full_transf = self.__create_compound_transf(from_frame, to_frame)

        return full_transf.apply(input)


    def wrench_change_frame(self, wrench: np.ndarray, from_frame: str, to_frame: str):
        # Verify input
        if not np.array(wrench).shape == (6,):
            log.error(f"TransformSet - Invalid wrench input. Shape must be (6,)")
            return

        # Create compound transformation
        full_transf = self.__create_compound_transf(from_frame, to_frame)

        # Transform wrench
        force_at_orig = wrench[:3]
        torque_at_orig = wrench[3:]

        torque_at_dest = full_transf.rotation.apply(np.cross(force_at_orig, full_transf.translation) + torque_at_orig)
        force_at_dest = full_transf.rotation.apply(force_at_orig)

        return np.hstack([force_at_dest, torque_at_dest])


    def wrench_df_change_frame(self, wrench_df: pd.DataFrame, from_frame: str, to_frame: str) -> pd.DataFrame:
        '''
        Function to perform coordinate transformation on wrench data in pandas dataframe. Returned dataframe
        will have the same index, column names and dimensions as the input dataframe.

        Note: Dataframe must have 6 columns, where the first three represent the forces on x, y, z; and the
            last three representing the torques on x, y, z (in that order).

        Args
            wrench_df (pd.DataFrame): Pandas dataframe containing wrench data in original frame
            from_frame (str): Name of origin frame
            to_frame (str): Name of target frame

        Returns
            pd.DataFrame: Dataframe containing wrench data in target frame (same index, column names and shape as input `wrench_df`)
        '''
        # Verify input
        if wrench_df.shape[1] != 6:
            log.error(f'TransformSet - Invalid wrench input. Dataframe should have 6 columns (given {wrench_df.shape[1]})')
            return

        # Create transformation matrix
        transf_mat = self.transform_matrix(from_frame=from_frame, to_frame=to_frame)

        # Get column names from dataframe and create output dataframe
        cols = wrench_df.columns
        transf_wrench = pd.DataFrame(index=wrench_df.index, columns=cols)

        # Compute unrotated moment
        unrot_moment = pd.DataFrame(index=wrench_df.index, columns=['Tx', 'Ty', 'Tz'])
        unrot_moment['Tx'] = wrench_df[cols[1]] * transf_mat[2, 3] - wrench_df[cols[2]] * transf_mat[1, 3] + wrench_df[cols[3]]
        unrot_moment['Ty'] = wrench_df[cols[2]] * transf_mat[0, 3] - wrench_df[cols[0]] * transf_mat[2, 3] + wrench_df[cols[4]]
        unrot_moment['Tz'] = wrench_df[cols[0]] * transf_mat[1, 3] - wrench_df[cols[1]] * transf_mat[0, 3] + wrench_df[cols[5]]

        # Perform coordinate transformation
        for i in range(len(cols)):
            if i < 3:
                transf_wrench[cols[i]] = transf_mat[i, 0]*wrench_df[cols[0]] + transf_mat[i, 1]*wrench_df[cols[1]] + transf_mat[i, 2]*wrench_df[cols[2]]
            else:
                transf_wrench[cols[i]] = transf_mat[i-3, 0]*unrot_moment['Tx'] + transf_mat[i-3, 1]*unrot_moment['Ty'] + transf_mat[i-3, 2]*unrot_moment['Tz']

        return transf_wrench


    def transform_matrix(self, from_frame: str, to_frame: str, homogeneous: bool = True):

        # Create compound transformation
        full_transf = self.__create_compound_transf(from_frame, to_frame)

        return full_transf.matrix()


    def __create_compound_transf(self, from_frame: str, to_frame: str) -> Transform:
        # Verify frame names
        if from_frame not in self.frame_names or to_frame not in self.frame_names:
            log.error(f"TransformSet - Invalid frame name, names must be: {self.frame_names}")
            return
        
        # Create transform to base frame from orig_frame
        if from_frame == 'base':
            transf_to_base = Transform(name="orig_is_base", orig='base', dest='base')
        else:
            orig_transf = self.transformations[self.frame_names.index(from_frame)]
            transf_to_base = orig_transf.inv()

        # Create transform from base frame from dest_frame
        if to_frame == 'base':
            transf_to_dest = Transform(name="dest_is_base", orig='base', dest='base')
        else:
            transf_to_dest = self.transformations[self.frame_names.index(to_frame)]

        # Create compound transformation
        full_transformation = Transform(name=f"{from_frame}2{to_frame}", orig=from_frame, dest=to_frame)
        full_transformation.translation = transf_to_base.rotation.apply(transf_to_dest.translation) + transf_to_base.translation
        full_transformation.rotation = Rotation.from_matrix(np.matmul(transf_to_base.rotation.as_matrix(), transf_to_dest.rotation.as_matrix()))

        return full_transformation



import numpy as np
from scipy.spatial.transform import Rotation

from pose3d.transform import Transform
from pose3d.pose import Pose
from pose3d.pose_comparison import PoseComparison


# Verify compatibility with poses
def pose_test(passing_criteria: float = 0.001):

    transf = Transform(name='demo')
    transf.random()

    inv_transf = transf.inv()

    pose = Pose(name='test')
    pose.random()

    result_1 = transf.apply(pose)
    result_2 = inv_transf.apply(result_1)

    position_difference, rotation_difference = PoseComparison.calc_difference(pose, result_2)
    assert np.linalg.norm(position_difference) < passing_criteria
    assert rotation_difference<passing_criteria

    return True


# Verify matrix generation
def transf_matrix_test(passing_criteria: float = 0.001):
    transf = Transform(name='demo')
    transf.random()

    inv_transf = transf.inv()

    mult = [np.matmul(transf.matrix(), inv_transf.matrix()),
            np.matmul(inv_transf.matrix(), transf.matrix())]

    for matrix in mult:
        for i in range(matrix.shape[0]):
            for j in range(matrix.shape[1]):
                assert abs(matrix[i, j] - np.eye(4)[i, j]) < passing_criteria

    return True


# Verify inverse calculation
def inv_test(passing_criteria: float = 0.001):

    transf = Transform(name='demo')
    transf.translation = np.random.rand(3)
    transf.rotation = Rotation.random()

    inv_transf = transf.inv()

    input_vec = np.random.rand(3)
    result_1 = transf.apply(input_vec)
    result_2 = inv_transf.apply(result_1)

    assert np.linalg.norm(np.subtract(result_2, input_vec)) < passing_criteria

    return True


def main():

    pass_criteria = 1e-10

    test_passed = inv_test(passing_criteria=pass_criteria)
    if test_passed:
        print("Test Passed: inv_test")

    test_passed = transf_matrix_test(passing_criteria=pass_criteria)
    if test_passed:
        print("Test Passed: transf_matrix_test")

    test_passed = pose_test(passing_criteria=pass_criteria)
    if test_passed:
        print("Test Passed: pose_test")

if __name__ == "__main__":
    main()



import numpy as np
from scipy.spatial.transform import Rotation


class PoseComparison:
    @staticmethod
    def different_pose(pose_1, pose_2, passing_criteria: float = 1e-5) -> bool:

        position_difference, rotation_difference = PoseComparison.calc_difference(pose_1, pose_2)

        if np.linalg.norm(position_difference) < passing_criteria and rotation_difference < passing_criteria:
            return True

        return False


    @staticmethod
    def calc_difference(pose_1, pose_2, degrees: bool = False):
        position_difference = np.subtract(pose_1.position, pose_2.position)

        if position_difference.shape == (2,):
            rotation_difference = abs(pose_1.orientation.as_euler(degrees) - pose_2.orientation.as_euler(degrees))

        elif position_difference.shape == (3,):
            mat = np.matmul(pose_1.orientation.inv().as_matrix(), pose_2.orientation.as_matrix())
            compound_rot = Rotation.from_matrix(mat)
            rotation_difference = np.linalg.norm(compound_rot.as_rotvec(degrees=degrees))

        return position_difference, rotation_difference


import numpy as np
from scipy.spatial.transform import Rotation

from .pose import Pose


class Transform:
    def __init__(self, name: str, orig: str = None, dest: str = None) -> None:
        # Set strings
        self.name = name

        if orig is None and dest is None:
            self.orig = 'origin'
            self.dest = 'destination'
        else:
            self.orig = orig
            self.dest = dest

        # Init translation and orientation
        self.translation = np.zeros(3)
        self.rotation = Rotation.identity()

    def print(self):
        print(f"Transformation: {self.name.title()}")
        print(f"Translation: {self.translation} [m]")
        print(f"Rotation:    {self.rotation.as_euler('xyz', degrees=True)} [deg]\n")

    def inv(self):
        inv_transform = Transform(name=f"{self.name} (Inverse)")
        inv_transform.rotation = self.rotation.inv()
        inv_transform.translation = -inv_transform.rotation.apply(self.translation)

        return inv_transform

    def apply(self, input):
        
        # If input is pose
        if type(input) is Pose:
            input.orientation.from_matrix(np.matmul(self.rotation.as_matrix(), input.orientation.as_matrix()))
            input.position += self.translation

        # If input is 3D vector
        elif type(input) is np.ndarray and np.size(input) == 3:
            input = self.rotation.apply(input) + self.translation

        return input

    def matrix(self, homogeneous: bool = True):

        matrix = np.eye(4)
        
        if self.orig != self.dest:
            matrix[:3, :3] = self.rotation.as_matrix()
            matrix[:3, 3] = self.translation

        if not homogeneous:
            return matrix[:3, :]
        
        return matrix

    def random(self):
        self.translation = np.random.rand(3)
        self.rotation = Rotation.random()



from .pose import Pose
from .pose_comparison import PoseComparison
from .pose2d import Pose2D
from .rotation2d import Rotation2D
from .transform_set import TransformSet
from .transform import Transform



import numpy as np
import logging as log
from math import pi, cos, sin
from random import uniform


class Rotation2D:
    def __init__(self) -> None:
        self.angle: float = 0.0    # Value is stored in radians

    def apply(self, input_vector: np.array) -> np.ndarray:
        if not np.size(input_vector) == 2:
            log.error(f"Rotation2D - Incorrect input size: {np.size(input_vector)}. Vector must be dimension 2.")
            return

        return np.matmul(self.as_matrix(), input_vector)

    def inv(self):
        self.angle = -self.angle

    def from_euler(self, value: float, degrees: bool):
        if degrees:
            self.angle = np.deg2rad(value)
        else:
            self.angle = value

    def from_matrix(self, matrix: np.ndarray):
        dummy_vector = np.array([1.0, 0.0])
        self.align_vectors(dummy_vector, np.matmul(matrix, dummy_vector))

    def as_euler(self, degrees: bool = False) -> float:
        if degrees:
            return np.rad2deg(self.angle)
        else:
            return self.angle

    def as_matrix(self) -> np.ndarray:
        matrix = np.array([[cos(self.angle), -sin(self.angle)],
                           [sin(self.angle), cos(self.angle)]])

        return matrix

    def identity(self):
        self.angle = 0.0

    def random(self):
        self.angle = uniform((0.0, 2.0*pi))

    def align_vectors(self, a: np.ndarray, b: np.ndarray):
        inner = np.inner(a, b)
        norms = np.linalg.norm(a) * np.linalg.norm(b)

        cos = inner / norms
        self.angle = np.arccos(np.clip(cos, -1.0, 1.0))     # Compute angle in radians



import numpy as np
from .rotation2d import Rotation2D

class Pose2D:
    def __init__(self, name) -> None:
        self.name = name
        self.orientation = Rotation2D()
        self.position = np.array([0.0, 0.0])

    def print(self):
        print(f"Pose2D: {self.name.title()}")
        print(f"Position: {self.position} [m]")
        print(f"Rotation: {self.orientation.as_euler(degrees=True)} [deg]\n")

    def random(self):
        self.orientation = Rotation2D.random()
        self.position = np.random.rand(2)



import numpy as np
from scipy.spatial.transform import Rotation

class Pose:
    def __init__(self, name: str) -> None:
        self.name = name
        self.orientation = Rotation.identity()
        self.position = np.array([0.0, 0.0, 0.0])

    def print(self):
        print(f"Pose: {self.name.title()}")
        print(f"Position:    {self.position} [m]")
        print(f"Orientation: {self.orientation.as_euler('xyz', degrees=True)} [deg]\n")

    def random(self):
        self.orientation = Rotation.random()
        self.position = np.random.rand(3)



from pose3d.rotation2d import Rotation2D

def euler2euler_test(angle: float, passing_criteria: float = 0.001):

    rotation = Rotation2D()
    rotation.from_euler(angle, degrees=True)

    assert (abs(angle - rotation.as_euler(degrees=True))/angle) < passing_criteria

    return True

def inv_test(angle: float, passing_criteria: float = 0.001):

    rotation = Rotation2D()
    rotation.from_euler(angle, degrees=True)
    rotation.inv()

    assert (abs(-angle - rotation.as_euler(degrees=True))/angle) < passing_criteria

    return True

def main():
    angle = 30

    test_passed = euler2euler_test(angle)
    if test_passed:
        print("euler2euler_test - Test Passed.")

    test_passed = inv_test(angle)
    if test_passed:
        print("inv_test - Test Passed.")

if __name__ == "__main__":
    main()
